import meta.Tuple;
import ctParams;


const float nearPlane	= 1.f;
const float aspect		= cast(float)imgWidth / imgHeight;

const float infinity	= 100000.f;


template sqrt(float x, int k = 5) {
	static if (k == 0) {
		const float sqrt = 1.f;
	} else {
		const float sqrt = .5f * (sqrt!(x, k-1) + x / sqrt!(x, k-1));
	}
}

static assert (sqrt!(0.04f) - 0.2f < 0.01f && sqrt!(0.04f) - 0.2f > -0.01f);



template isVec3(T) {
	static if (is(typeof(T.x)))
	static if (is(typeof(T.y)))
	static if (is(typeof(T.z)))
	static if (is(T == Vec3!(T.x, T.y, T.z)))
		const bool isVec3 = true;
	else const bool isVec3 = false;
	else const bool isVec3 = false;
	else const bool isVec3 = false;
	else const bool isVec3 = false;
}



struct Vec3(float x_, float y_, float z_) {
	alias Vec3 _;

	const float x = x_;
	const float y = y_;
	const float z = z_;
}


template dot(X, T) {
	static assert(isVec3!(X));
	static assert(isVec3!(T));
	const float dot = X.x * T.x + X.y * T.y + X.z * T.z;
}


template subtract(X, T) {
	static assert(isVec3!(X));
	static assert(isVec3!(T));
	alias Vec3!(X.x - T.x, X.y - T.y, X.z - T.z) subtract;
}


template add(X, T) {
	static assert(isVec3!(X));
	static assert(isVec3!(T));
	alias Vec3!(X.x + T.x, X.y + T.y, X.z + T.z) add;
}


template multiply(X, T) {
	static assert(isVec3!(X));
	static assert(isVec3!(T));
	alias Vec3!(X.x * T.x, X.y * T.y, X.z * T.z) multiply;
}


template scale(X, float s) {
	static assert(isVec3!(X));
	alias Vec3!(X.x * s, X.y * s, X.z * s) scale;
}



template length(X) {
	static assert(isVec3!(X));
	const float length = sqrt!(dot!(X, X));
}


template cross(A, B) {
	static assert(isVec3!(A));
	static assert(isVec3!(B));
	alias Vec3!(
			A.y * B.z - B.y * A.z,
			A.z * B.x - B.z * A.x,
			A.x * B.y - B.x * A.y
		) cross;	
}



struct Ray(origin_, dir_) {
	alias Ray _;

	static assert (isVec3!(origin_));
	static assert (isVec3!(dir_));

	alias origin_	origin;
	alias dir_		dir;
}




struct Sphere(center_, float radius_, color_) {
	alias Sphere _;

	static assert (isVec3!(center_));
	static assert (isVec3!(color_));

	alias center_	center;
	alias color_	color;
	const float		radius = radius_;


	template intersect(ray) {
		alias subtract!(ray.origin, center) dst;
		const float r2 = radius * radius;
		const float B = dot!(dst, ray.dir);
		const float C = dot!(dst, dst) - r2;
		const float D = B*B - C;

		static if (D > 0.f) {
			const float distance = -B -sqrt!(D);
		} else {
			const float distance = infinity;
		}


		static if (distance >= infinity || distance < 0.f) {
			alias void result;
		} else {
			alias scale!(ray.dir, distance) offset;
			alias add!(ray.origin, offset) POI;

			alias subtract!(POI, center) norm_;
			alias scale!(norm_, 1.f / radius) normal;

			alias Intersection!(distance, POI, normal, color) result;
		}
	}
}


struct Plane(normal_, float D_, color_) {
	alias Plane _;

	static assert (isVec3!(normal_));
	static assert (isVec3!(color_));

	alias normal_	normal;
	alias color_	color;
	const float		D = D_;


	template intersect(ray) {
		const float d = dot!(normal, ray.dir);
		static if (d <= -0.001f)
		{
			const float dist = -(dot!(normal, ray.origin) + D) / d;
			static if (dist > 0.f)
			{
				alias scale!(ray.dir, dist) offset;
				alias add!(ray.origin, offset) POI;

				alias Intersection!(dist, POI, normal, color) result;
			} else {
				alias void result;
			}
		} else {
			alias void result;
		}
	}
}



struct Triangle(v0_, v1_, v2_, color_) {
	alias Triangle _;

	static assert (isVec3!(v0_));
	static assert (isVec3!(v1_));
	static assert (isVec3!(v2_));
	static assert (isVec3!(color_));

	alias v0_		v0;
	alias v1_		v1;
	alias v2_		v2;
	alias color_	color;

	alias subtract!(v1, v0) edge1;
	alias subtract!(v2, v0) edge2;

	alias scale!(cross!(edge1, edge2), 1.f / length!(cross!(edge1, edge2))) normal;


	template intersect(ray) {
		const float e = 0.001f;
		
		
		/+ begin calculating determinant - also used to calculate U parameter +/
		alias cross!(ray.dir, edge2) pvec;
		
		/+ if determinant is near zero, ray lies in plane of triangle +/
		const float det = dot!(edge1, pvec);
		
		static if (det < e) {
		   	alias void result;
		} else {
			/+ calculate distance from vert0 to ray origin +/
			alias subtract!(ray.origin, v0) tvec;
		
			/+ calculate U parameter and test bounds +/
			const float u = dot!(tvec, pvec);
			static if (u < 0.f || u > det) {
				alias void result;
			} else {
				/+ prepare to test V parameter +/
				alias cross!(tvec, edge1) qvec;
		
				/+ calculate V parameter and test bounds +/
				const float v = dot!(ray.dir, qvec);
				static if (v < 0.f || u + v > det) {
					alias void result;
				} else {
					/+ calculate t, ray intersects triangle +/
					const float t_ = dot!(edge2, qvec);
					const float inv_det = 1.f / det;
					const float t = t_ * inv_det;

					alias scale!(ray.dir, t) offset;
					alias add!(ray.origin, offset) POI;
					alias Intersection!(t, POI, normal, color) result;
				}
			}
		}
	}
}



struct Box(center_, extents_, color_) {
	alias Box _;

	static assert (isVec3!(center_));
	static assert (isVec3!(extents_));
	static assert (isVec3!(color_));

	alias center_	center;
	alias extents_	extents;
	alias color_	color;

	template intersect(ray) {
		alias subtract!(center, extents)	minB;
		alias add!(center, extents)			maxB;

		alias ray.origin	o;
		alias ray.dir		d;

		const float e = 0.001f;

		// check left - right collision
		static if ((o.x < minB.x+e && d.x > 0.f) || (o.x > maxB.x-e && d.x < 0.f)) {
			static if (d.x > 0.f) const float ix = minB.x;
			else const float ix = maxB.x;

			const float x1 = d.x;
			const float x2 = ix - o.x;

			alias Vec3!(ix, o.y + x2 * d.y / x1, o.z + x2 * d.z / x1) xpoi;
			static if (xpoi.y <= maxB.y+e && xpoi.y >= minB.y-e && xpoi.z <= maxB.z+e && xpoi.z >= minB.z-e) {
				alias xpoi poi;
				const float resultDist = x2 / x1;
				static if (d.x > 0.f) alias Vec3!(-1.f, 0.f, 0.f) norm;
				else alias Vec3!(1.f, 0.f, 0.f) norm;
			}
		}


		// check bottom-top collision
		static if (!is(typeof(resultDist)) && ((o.y < minB.y+e && d.y > 0.f) || (o.y > maxB.y-e && d.y < 0.f))) {
			static if (d.y > 0.f) const float iy = minB.y;
			else const float iy = maxB.y;

			const float y1 = d.y;
			const float y2 = iy - o.y;

			alias Vec3!(o.x + y2 * d.x / y1, iy, o.z + y2 * d.z / y1) ypoi;
			static if (ypoi.x <= maxB.x+e && ypoi.x >= minB.x-e && ypoi.z <= maxB.z+e && ypoi.z >= minB.z-e) {
				alias ypoi poi;
				const float resultDist = y2 / y1;
				static if (d.y > 0.f) alias Vec3!(0.f, -1.f, 0.f) norm;
				else alias Vec3!(0.f, 1.f, 0.f) norm;
			}
		}


		// check back-front collision
		static if (!is(typeof(resultDist)) && ((o.z < minB.z+e && d.z > 0.f) || (o.z > maxB.z-e && d.z < 0.f))) {
			static if (d.z > 0.f) const float iz = minB.z;
			else const float iz = maxB.z;

			const float z1 = d.z;
			const float z2 = iz - o.z;

			alias Vec3!(o.x + z2 * d.x / z1, o.y + z2 * d.y / z1, iz) zpoi;
			static if (zpoi.x <= maxB.x+e && zpoi.x >= minB.x-e && zpoi.y <= maxB.y+e && zpoi.y >= minB.y-e) {
				alias zpoi poi;
				const float resultDist = z2 / z1;
				static if (d.z > 0.f) alias Vec3!(0.f, 0.f, -1.f) norm;
				else alias Vec3!(0.f, 0.f, 1.f) norm;
			}
		}


		static if (is(typeof(resultDist))) {
			alias Intersection!(resultDist, poi, norm, color) result;
		} else {
			alias void result;
		}
	}
}


template cosineLightFalloff(ptNormal, lightDir) {
	const float dotRes = -dot!(ptNormal, lightDir);

	static if (dotRes < 0.f) {
		const float result = 0.f;
	} else static if (dotRes > 1.f) {
		const float result = 1.f;
	} else {
		const float result = dotRes;
	}
}


struct DirectionalLight(dir_, color_, bool castsShadows_ = false) {
	alias DirectionalLight _;

	static assert (isVec3!(dir_));
	static assert (isVec3!(color_));

	alias scale!(dir_, 1.f / length!(dir_))		dir;
	alias color_								color;

	const bool castsShadows = castsShadows_;


	static template getPointLighting(pos, normal) {
		static if (castsShadows) {
			alias subtract!(Vec3!(0.f, 0.f, 0.f), dir) shadowRayDir;
			alias Ray!(pos, shadowRayDir) shadowRay;
			alias intersectRayScene!(shadowRay, scene).result isect;
		} else alias void isect;

		static if (is(isect == void)) {
			const float val = cosineLightFalloff!(normal, dir).result;
			alias scale!(color, val) result;
		} else {
			alias Vec3!(0.f, 0.f, 0.f) result;
		}
	}
}



struct PointLight(pos_, color_) {
	alias PointLight _;

	static assert (isVec3!(pos_));
	static assert (isVec3!(color_));

	alias pos_		pos;
	alias color_	color;


	static template getPointLighting(pt, normal) {
		const float dist = length!(subtract!(pt, pos));

		alias subtract!(pos, pt) shadowRayDir;
		alias Ray!(pt, scale!(shadowRayDir, 1.f / length!(shadowRayDir))) shadowRay;
		alias intersectRayScene!(shadowRay, scene).result isect;

		static if (is(isect == void)) {
			const float ldist = infinity;
		} else const float ldist = isect.dist;

		static if (ldist >= dist) {
			alias scale!(shadowRayDir, -1.f / length!(shadowRayDir)) lightDir;
			const float val = cosineLightFalloff!(normal, lightDir).result / (1.f + dist);
			alias scale!(color, val) result;
		} else {
			alias Vec3!(0.f, 0.f, 0.f) result;
		}
	}
}



struct Intersection(float dist_, pos_, normal_, color_) {
	static assert (isVec3!(pos_));
	static assert (isVec3!(normal_));
	static assert (isVec3!(color_));

	const float dist = dist_;

	alias pos_		pos;
	alias normal_	normal;
	alias color_	color;
}


template valToString(float i) {
	static if (i > 1.f) const float x = 1.f;
	else static if (i < 0.f) const float x = 0.f;
	else const float x = i;

	const char result[] = "\x01\x01\x02\x03\x04\x05\x06\x06\x06\x09\x09\x0b\x0c\x0c\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f\x20\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2a\x2b\x2c\x2d\x2e\x2f\x30\x31\x32\x33\x34\x35\x36\x37\x38\x39\x3a\x3b\x3c\x3d\x3e\x3f\x40\x41\x42\x43\x44\x45\x46\x47\x48\x49\x4a\x4b\x4c\x4d\x4e\x4f\x50\x51\x52\x53\x54\x55\x56\x57\x58\x59\x5a\x5b\x5c\x5d\x5e\x5f\x60\x61\x62\x63\x64\x65\x66\x67\x68\x69\x6a\x6b\x6c\x6d\x6e\x6f\x70\x71\x72\x73\x74\x75\x76\x77\x78\x79\x7a\x7b\x7c\x7d\x7e\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff"
		[cast(ubyte)(255 * x)] ~ "";
}


template colorToString_binary(T) {
	static assert(isVec3!(T));
	const char[] colorToString_binary = valToString!(T.z).result ~ valToString!(T.y).result ~ valToString!(T.x).result;
}






alias Tuple!(
		Sphere!(Vec3!(-4.f, 1.5f, -8.f),    3.2f,  Vec3!(1.f, .3f, .1f)),
		Sphere!(Vec3!(2.5f,  1.5f, -5.0f),  1.2f,  Vec3!(.2f, 1.f, .4f)),
		Sphere!(Vec3!( .7f, -1.5f, -4.2f),  1.5f,  Vec3!(.2f, .4f, 1.f)),

		//Sphere!(Vec3!( .0f, .0f, -2.f),  1.f,  Vec3!(.2f, .4f, 1.f))
		//Box   !(Vec3!( -4.f, 1.5f, -8.f),  Vec3!(1.f, 3.f, 3.f),  Vec3!(1.f, .2f, .2f)),
		//Sphere!(Vec3!(-.6f, .0f, -2.f),  .2f,  Vec3!(.2f, .4f, 1.f)),

		Plane! (Vec3!( 0.f,  1.f,   0.f),   4.f,   Vec3!(.5f, .5f, .5f))
) scene;


alias Tuple!(
		//DirectionalLight!(Vec3!(-0.4f, -0.4f, -0.7f), Vec3!(0.7f, 0.7f, 0.6f)),
		PointLight!(Vec3!(8.f, 8.f, 10.f), scale!(Vec3!(0.7f, 0.7f, 0.6f), 30.f)),
		//DirectionalLight!(Vec3!( 0.6f,  0.7f, 0.2f), Vec3!(.2f, .2f, .4f), false),
		PointLight!(Vec3!(-2.f,  2.5f, 0.f), scale!(Vec3!(.6f, .4f, .1f), 8.f))
) lights;


alias colorToString_binary colorToString;
const char[] newline = "";
typedef void less;
typedef void greater;



template intersectRayScene(ray, scene, cmpFunc = less, int i = 0) {
	static assert ((is(cmpFunc == less) && !is(cmpFunc == greater)) || (!is(cmpFunc == less) && is(cmpFunc == greater)));

	static if (i < scene.length) {
		alias typeof(scene.mix.val!(i)) obj;

		alias obj._.intersect!(ray).result isect;

		static if (i+1 < scene.length) {
			alias .intersectRayScene!(ray, scene, cmpFunc, i+1).result isect2;

			static if (is(isect2 == void)) {
				alias isect result;
			} else static if (is(isect == void)) {
				alias isect2 result;
			} else static if ((is(cmpFunc == less) && isect2.dist < isect.dist) || (is(cmpFunc == greater) && isect2.dist > isect.dist)) {
				alias isect2 result;
			} else {
				alias isect result;
			}
		} else alias isect result;
	} else {
		alias void result;
	}
}


// i - light number
template getPointLighting(pos, normal, int i = 0) {
	static assert (isVec3!(pos));
	static assert (isVec3!(normal));

	static if (i < lights.length) {
		alias lights.mix.val!(i)._.getPointLighting!(pos, normal).result thisl;
		alias add!(.getPointLighting!(pos, normal, i+1).result, thisl) result;
	} else {
		alias Vec3!(0.f, 0.f, 0.f) result;
	}
}


template getReflection(ray, normal) {
	static assert (isVec3!(ray));
	static assert (isVec3!(normal));

	alias subtract!(ray, scale!(normal, 2.f * dot!(ray, normal))) getReflection;
}



const int maxReflections = 2;

template traceRayFull(ray, scene, int i=0) {
	alias intersectRayScene!(ray, scene).result isect;

	static if (is(isect == void)) {
		alias Vec3!(0.f, 0.f, 0.f) result;
	} else {
		alias getPointLighting!(isect.pos, isect.normal).result light;
		alias multiply!(isect.color, light) col1;

		static if (i < maxReflections) {
			alias getReflection!(ray.dir, isect.normal) rdir;
			alias add!(isect.pos, scale!(rdir, 0.001f)) rorigin;

			alias traceRayFull!(Ray!(rorigin, rdir), scene, i+1).result col2;
			alias add!(scale!(col1, 0.7f), scale!(col2, 0.3f)) result;
		} else {
			alias col1 result;
		}
	}
}


template renderScanline(int hi, int wi = 0) {
	static if (wi-xOffset < spanLen) {
		const float x = (nearPlane * 2.f * cast(float)wi / imgWidth   - 1.f) * aspect;
		const float y =  nearPlane * 2.f * cast(float)hi / imgHeight  - 1.f;
		const float z = -nearPlane;
		
		alias Vec3!(x, y, z) direction;
		alias scale!(direction, 1.f / length!(direction)) normDir;

		alias traceRayFull!(Ray!(Vec3!(0.f, 0.f, 0.f), normDir), scene).result pxColor;
		const char[] px = colorToString!(pxColor);

		static if (wi+1-xOffset < spanLen) {
			const char[] result = px ~ .renderScanline!(hi, wi+1).result;
		} else {
			alias px result;
		}
	} else {
		char[] result = "";
	}
}


pragma (msg, renderScanline!(yOffset, xOffset).result);

